@@ -188,7 +188,7 @@ keybind (cycle_panels_backward,    handle_cycle,         META_TAB_LIST_DOCKS,
 /***********************************/
      
 keybind (show_desktop, handle_show_desktop, 0, 0, "<Control><Alt>d",
-      _("Hide all normal windows and set focus to the desktop background"))
+      _("Hide all normal windows and set focus to the desktop"))
 keybind (panel_main_menu, handle_panel,
        META_KEYBINDING_ACTION_PANEL_MAIN_MENU, 0, "<Alt>F1",
       _("Show the panel's main menu"))
@@ -258,14 +258,14 @@ keybind (activate_window_menu, handle_activate_window_menu, 0,
 keybind (toggle_fullscreen, handle_toggle_fullscreen, 0, BINDING_PER_WINDOW,
         NULL,
         _("Toggle fullscreen mode"))
-keybind (toggle_maximized, handle_toggle_maximized, 0, BINDING_PER_WINDOW, NULL,
+keybind (toggle_maximized, handle_toggle_maximized, 0, BINDING_PER_WINDOW, "<Alt>F10",
         _("Toggle maximization state"))
 keybind (toggle_above, handle_toggle_above, 0, BINDING_PER_WINDOW, NULL,
         _("Toggle whether a window will always be visible over other windows"))
-keybind (maximize, handle_maximize, 0, BINDING_PER_WINDOW, "<Alt>F10",
+keybind (maximize, handle_maximize, 0, BINDING_PER_WINDOW, NULL,
         _("Maximize window"))
 keybind (unmaximize, handle_unmaximize, 0, BINDING_PER_WINDOW, "<Alt>F5",
-        _("Unmaximize window"))
+        _("Restore window"))
 keybind (toggle_shaded, handle_toggle_shaded, 0, BINDING_PER_WINDOW, NULL,
         _("Toggle shaded state"))
 keybind (minimize, handle_minimize, 0, BINDING_PER_WINDOW, "<Alt>F9",
@@ -362,7 +362,7 @@ keybind (move_to_corner_ne, handle_move_to_corner_ne, 0,
 keybind (move_to_corner_sw, handle_move_to_corner_sw, 0,
         BINDING_PER_WINDOW, NULL,
         _("Move window to south-west (bottom left) corner"))
-keybind (move_to_corner_nw, handle_move_to_corner_se, 0,
+keybind (move_to_corner_se, handle_move_to_corner_se, 0,
         BINDING_PER_WINDOW, NULL,
         _("Move window to south-east (bottom right) corner"))
 
